#!/usr/bin/env python
# encoding: utf-8
#
# Wellcome Trust Sanger Institute
# Copyright (C) 2013  Wellcome Trust Sanger Institute
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import sys
import os
import subprocess

from Bio import SeqIO

from gubbins import utils

class Star:
    """Class for constructing star phylogenies"""
    
    def __init__(self):
        self.executable = "star phylogeny"
        self.tree_prefix = ""
        self.tree_suffix = ".tre"
        self.alignment_suffix = ".snp_sites.aln"
        # Reproducibility
        self.name = "Star"
        self.model = "-"
        self.version = "unspecified"
        self.citation = "no citation"
        self.invariant_site_correction = False
    
    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        # Extract taxon names from alignment
        taxon_names = []
        with open(alignment_filename) as input_handle:
            for record in SeqIO.parse(input_handle, "fasta"):
                taxon_names.append(record.id)

        # Write tree
        star_tree_string = "("
        star_tree_string = star_tree_string + ':0.09,'.join(taxon_names)
        star_tree_string = star_tree_string + ':0.1);' # mid point rooting fails with equidistant taxa

        # Print to file
        output_tree = basename + self.tree_suffix
        with open(output_tree,'w') as out_file:
            out_file.write(star_tree_string + '\n')
        return output_tree

class RapidNJ:
    """Class for operations with the rapidNJ executable"""

    def __init__(self, threads = int, model='GTRCAT', bootstrap = 0, verbose=False, additional_args = None):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.tree_prefix = ""
        self.tree_suffix = ".tre"
        self.alignment_suffix = ".snp_sites.aln"
        self.model = model
        self.additional_args = additional_args
        self.bootstrap = bootstrap
        self.invariant_site_correction = False

        # Construct command
        self.executable = "rapidnj"
        if utils.which(self.executable) is None:
            sys.exit("No usable version of rapidnj could be found.")

        # Reproducibility
        self.name = 'RapidNJ'
        self.version = "unspecified"
        self.citation = "https://doi.org/10.1007/978-3-540-87361-7_10"

        command = [self.executable]
        command.extend(["-i fa", "-t d", "-n"])
        command.extend(["-c", str(self.threads)])
        if self.model == 'JC':
            command.extend(["-a", "jc"])
        elif self.model == 'K2P':
            command.extend(["-a", "kim"])
        else:
            command.extend(["-a", self.model])
        # Additional arguments
        if self.additional_args is not None:
            command.extend([self.additional_args])
        self.base_command = command

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the rapidNJ executable"""
        command = self.base_command.copy()
        # Alignment file needs to be first argument
        executable = command.pop(0)
        command.insert(0,alignment_filename)
        command.insert(0,executable)
        # Specify output file
        output_tree = basename + self.tree_suffix
        command.extend(["-x", output_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)
        
    def bootstrapping_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Runs a bootstrapping analysis and annotates the nodes of a summary tree"""
        command = self.base_command.copy()
        # Alignment file needs to be first argument
        executable = command.pop(0)
        command.insert(0,alignment_filename)
        command.insert(0,executable)
        # Specify output file
        output_tree = basename + self.tree_suffix + '.bootstrapped'
        # Number of bootstraps
        command.extend(["-b", str(self.bootstrap)])
        command.extend(["-x", output_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

class FastTree:
    """Class for operations with the FastTree executable"""

    def __init__(self, threads: int, bootstrap = 0, model='GTRCAT', seed = None, verbose=False, additional_args = None):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.model = model
        self.tree_prefix = ""
        self.tree_suffix = ".tre"
        self.alignment_suffix = ".snp_sites.aln"
        self.bootstrap = bootstrap
        self.additional_args = additional_args
        self.seed = utils.set_seed(seed)
        self.invariant_site_correction = False

        # Identify executable
        self.potential_executables = ["FastTreeMP","fasttreeMP","FastTree", "fasttree"]
        self.executable = utils.choose_executable(self.potential_executables)
        if self.executable is None:
            sys.exit("No usable version of FastTree could be found.")

        # Reproducibility
        self.name = 'FastTree'
        self.version = self.get_version(self.executable)
        self.citation = "https://doi.org/10.1371/journal.pone.0009490"

        # Function for returning base command
        command = [self.executable]
        command.extend(["-nt"])
        command.extend(["-rawdist"]) # https://morgannprice.github.io/fasttree/
        if self.model == 'JC':
            command.extend(["-nocat"])
        elif self.model == 'GTR':
            command.extend(["-gtr","-nocat"])
        elif self.model == 'GTRGAMMA':
            command.extend(["-gtr","-gamma"])
        elif self.model == 'GTRCAT':
            command.extend(["-gtr"])
        else:
            command.extend([self.model])
        command.extend(["-seed",self.seed])
        # Additional arguments
        if self.additional_args is not None:
            command.extend([self.additional_args])
        self.base_command = command
        
        # Set the number of threads for parallelisation
        omp_threads_command = 'export OMP_NUM_THREADS=' + str(self.threads)
        try:
            subprocess.check_call(omp_threads_command, shell=True)
        except subprocess.SubprocessError:
            sys.exit("Failed to set number of threads for fasttree with command " + omp_threads_command)

    def get_version(self,exe) -> str:
        """Gets the version of the tree building algorithm being used"""
        version = "Not determined"
        version_message = subprocess.run([exe], capture_output=True)
        for line in version_message.stderr.decode().splitlines():
            if line.startswith('Usage'):
                info = line.split()
                if len(info) >= 5:
                    version = info[4]
                break
        return version

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the FastTree executable"""
        command = self.base_command.copy()
        if input_tree:
            command.extend(["-intree", input_tree])
        output_tree = basename + self.tree_suffix
        command.extend(["-nosupport"])
        command.extend(["-out", output_tree])
        command.extend(["-log", basename + '.log'])
        command.append(alignment_filename)
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)
    
    def get_info_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the file containing the fitted model parameters"""
        fn = tmp + '/' + basename + '.log'
        return fn
    
    def get_recontree_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the tree generated by model fitting"""
        fn = tmp + '/' + basename + '.treefile'
        return fn

    def model_fitting_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Fits a nucleotide substitution model to a tree and an alignment"""
        command = self.base_command.copy()
        command.extend(["-mllen","-nome"])
        command.extend(["-nosupport"])
        command.extend(["-intree",input_tree])
        command.extend(["-log", basename + ".log"])
        command.extend(["-out", basename + ".treefile"])
        command.extend([alignment_filename])
        return " ".join(command)
        
    def bootstrapping_command(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a bootstrapping analysis and annotates the nodes of a summary tree"""
        command = self.base_command.copy()
        output_tree = basename + self.tree_suffix
        command.extend(["-nosupport"])
        command.extend(["-intree1",input_tree]) # http://www.microbesonline.org/fasttree/
        command.extend(["-out", tmp + "/" + basename + ".bootstrapped_trees"])
        command.extend(["-log", basename + ".log"])
        command.extend(["-n", str(self.bootstrap)])
        command.append(alignment_filename + ".bootstrapping.aln")
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)
    
    def sh_test(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a single branch support test"""
        command = self.base_command.copy()
        command.extend(["-mllen","-nome"])
        command.extend(["-intree",input_tree])
        command.extend(["-out",tmp + "/" + input_tree + ".sh_support"])
        command.extend([alignment_filename])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def get_bootstrapped_trees_file(self, tmp: str, basename: str) -> str:
        """Return bootstrapped tree files name"""
        file_name = tmp + "/" + basename + ".bootstrapped_trees"
        return file_name

class IQTree:
    """Class for operations with the IQTree executable"""

    def __init__(self, threads: 1, model: str, bootstrap = 0, invariant_proportion = 0, constant_base_counts = [0.0,0.0,0.0,0.0], seed = None, internal_node_prefix="", verbose=False, use_best=False, invariant_site_correction = True, additional_args = None):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.model = model
        self.tree_prefix = ""
        self.tree_suffix = ".treefile"
        self.asr_prefix = ""
        self.asr_suffix = ".state"
        self.asr_tree_prefix = ""
        self.asr_tree_suffix = ".treefile"
        self.alignment_suffix = ".phylip"
        self.internal_node_prefix = internal_node_prefix
        self.bootstrap = bootstrap
        self.use_best = use_best
        self.seed = utils.set_seed(seed)
        self.additional_args = additional_args
        self.invariant_site_correction = (invariant_site_correction if max(constant_base_counts) > 0 else False)
    
        # Construct base command
        self.executable = "iqtree"
        if utils.which(self.executable) is None:
            sys.exit("No usable version of IQTree could be found.")
        command = [self.executable]

        # Reproducibility
        self.name = 'IQTree'
        self.version = self.get_version(self.executable)
        self.citation = "https://doi.org/10.1093/molbev/msaa015"

        # Set parallelisation
        command.extend(["-T", str(self.threads)])

        # Define model
        command.extend(["-safe","-redo"])
        if self.use_best:
            pass
        else:
            invariant_site_string = ""
            if invariant_site_correction:
                #invariant_site_string = "+I{" + str(invariant_proportion) + "}"
                invariant_site_string = "+I"
            if self.model == 'JC':
                command.extend(["-m", "JC"+invariant_site_string])
            elif self.model == 'K2P':
                command.extend(["-m", "K2P"+invariant_site_string])
            elif self.model == 'HKY':
                command.extend(["-m", "HKY"+invariant_site_string])
            elif self.model == 'GTR':
                command.extend(["-m","GTR"+invariant_site_string])
            elif self.model == 'GTRGAMMA':
                command.extend(["-m","GTR+G4"+invariant_site_string])
            else:
                command.extend(["-m",self.model])
        command.extend(["-seed",self.seed])
        
        # Account for invariant sites
        if invariant_site_correction:
            command.extend(["-fconst",','.join([str(x) for x in constant_base_counts])])
        
        # Additional arguments
        if self.additional_args is not None:
            command.extend([self.additional_args])
        self.base_command = command

    def get_version(self,exe) -> str:
        """Gets the version of the tree building algorithm being used"""
        version = "Not determined"
        version_message = subprocess.run([exe], capture_output=True)
        for line in version_message.stdout.decode().splitlines():
            if line.startswith('IQ-TREE'):
                info = line.split()
                if len(info) >= 4:
                    version = info[3]
                break
        return version

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the IQTree executable"""
        command = self.base_command.copy()
        command.extend(["-s", alignment_filename, "-pre", basename])
        if input_tree:
            command.extend(["-t", input_tree])
        if not self.verbose:
            command.append("-quiet")
        return " ".join(command)

    def internal_sequence_reconstruction_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the IQTree executable for ancestral sequence reconstruction"""
        command = self.base_command.copy()
        command.extend(["-asr"])
        command.extend(["-s", alignment_filename, "-pre", basename])
        if input_tree:
            command.extend(["-te", input_tree])
        if not self.verbose:
            command.append("-quiet")
        return " ".join(command)

    def convert_raw_ancestral_states_to_fasta(self, input_filename, output_filename):
        """Converts the file containing ancestral sequences into FASTA format"""
        raw_sequences = {}
        with open(input_filename, 'r') as infile:
            for line in infile:
                if not line.startswith("#") and not line.startswith("Node\tSite\tState"):
                    elements = line.split("\t")
                    name = elements[0]
                    base = elements[2]
                    if name in raw_sequences:
                        raw_sequences[name].append(base)
                    else:
                        raw_sequences[name] = [base]

        with open(output_filename, 'w+') as outfile:
            for sequence_name, sequence_bases in raw_sequences.items():
                outfile.write(">" + self.replace_internal_node_label(sequence_name) + "\n")
                outfile.write("".join(sequence_bases) + "\n")

    def replace_internal_node_label(self, label):
        """Changes the label of internal nodes"""
        return self.internal_node_prefix + label.replace("Node", "")

    def get_info_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the file containing the fitted model parameters"""
        fn = tmp + '/' + basename + '.iqtree'
        return fn
    
    def get_recontree_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the tree generated by model fitting"""
        fn = tmp + '/' + basename + '.treefile'
        return fn

    def model_fitting_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Fits a nucleotide substitution model to a tree and an alignment"""
        # Using http://www.iqtree.org/doc/Advanced-Tutorial#user-defined-substitution-models
        command = self.base_command.copy()
        command.extend(["-s", alignment_filename, "-t", input_tree, "--prefix", basename, " -n 0 --mlrate", "-redo"])
        return " ".join(command)
    
    def bootstrapping_command(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a bootstrapping analysis"""
        command = self.base_command.copy()
        command.extend(["-s", alignment_filename, "-t", input_tree, "--prefix", tmp + "/" + basename + ".bootstrapped", "-B", str(self.bootstrap), "-wbt"])
        return " ".join(command)

    def sh_test(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a single branch support test"""
        command = self.base_command.copy()
        command.extend(["-s", alignment_filename])
        command.extend(["--prefix", tmp + "/" + input_tree + ".sh_support"])
        command.extend(["-te", input_tree])
        command.extend(["-alrt 0"])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)
        
    def get_bootstrapped_trees_file(self, tmp: str, basename: str) -> str:
        """Return bootstrapped tree files name"""
        file_name = tmp + "/" + basename + ".bootstrapped.ufboot"
        return file_name
    
    def run_time_tree(self, alignment_filename: str, input_tree: str, date_file: str, tmp: str, basename: str, outgroup=None) -> str:
        """Run time calibration of tree"""
        command = self.base_command.copy()
        command.extend(["-s", alignment_filename])
        command.extend(["-te", input_tree, "--tree-fix"])
        command.extend(["--date", date_file])
        command.extend(["--prefix", os.path.join(tmp,basename)])
        command.extend(["-blfix"])
        command.extend(["--date-options","' -l 0'"])
        if outgroup is not None:
            command.extend(["-o", outgroup])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def run_model_comparison(self, alignment_filename: str, basename: str) -> str:
        """Pick best model based on ML fit to data"""
        command = self.base_command.copy()
        command.extend(["-s", alignment_filename])
        command.extend(["-m TESTONLY"])
        command.extend(["-mset JC,K2P,HKY,GTR -cmax 4"])
        command.extend(["--prefix",basename])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

class RAxML:
    """Class for operations with the RAxML executable"""

    def __init__(self, threads: 1, model='GTRCAT', constant_base_counts = [0,0,0,0], partition_length = 1, bootstrap = 0, seed = None, internal_node_prefix="", verbose=False, invariant_site_correction = True, additional_args = None):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.model = model
        self.tree_prefix = "RAxML_result."
        self.tree_suffix = ""
        self.asr_prefix = "RAxML_marginalAncestralStates."
        self.asr_suffix = ""
        self.asr_tree_prefix = "RAxML_nodeLabelledRootedTree."
        self.asr_tree_suffix = ""
        self.alignment_suffix = ".phylip"
        self.internal_node_prefix = internal_node_prefix
        self.bootstrap = bootstrap
        self.seed = utils.set_seed(seed)
        self.constant_base_counts = constant_base_counts
        self.partition_length = partition_length
        self.additional_args = additional_args
        self.invariant_site_correction = (invariant_site_correction if max(constant_base_counts) > 0 else False)

        self.single_threaded_executables = ['raxmlHPC-AVX2', 'raxmlHPC-AVX', 'raxmlHPC-SSE3', 'raxmlHPC']
        self.multi_threaded_executables = ['raxmlHPC-PTHREADS-AVX2', 'raxmlHPC-PTHREADS-AVX',
                                           'raxmlHPC-PTHREADS-SSE3', 'raxmlHPC-PTHREADS']
        self.executable = self.select_executable_based_on_threads()
        if self.executable is None:
            sys.exit("No usable version of RAxML could be found.")
        command = [self.executable]
        
        # Reproducibility
        self.name = 'RAxML'
        self.version = self.get_version(self.executable)
        self.citation = "https://doi.org/10.1093/bioinformatics/btu033"

        # Set parallelisation
        if self.threads > 1:
            command.extend(["-T", str(self.threads)])

        # Add flags
        command.extend(["-safe"])
        invariant_site_string = ""
        model_string = "GTRGAMMA"
        if self.invariant_site_correction:
            model_string = "ASC_GTRGAMMA"
            invariant_site_string = "--asc-corr=stamatakis"
        if self.model == 'JC':
            command.extend(["-m", model_string,"--JC69",invariant_site_string])
        elif self.model == 'K2P':
            command.extend(["-m", model_string,"--K80",invariant_site_string])
        elif self.model == 'HKY':
            command.extend(["-m", model_string,"--HKY85",invariant_site_string])
        elif self.model == 'GTRGAMMA':
            command.extend(["-m", model_string,invariant_site_string])
        else:
            if self.model.startswith("ASC_"):
                command.extend(["-m", self.model,"--asc-corr=stamatakis"])
            else:
                self.invariant_site_correction = False
                command.extend(["-m", self.model])
        command.extend(["-p",self.seed])
        # Additional arguments
        if self.additional_args is not None:
            command.extend([self.additional_args])
        self.base_command = command

    def get_version(self,exe) -> str:
        """Gets the version of the tree building algorithm being used"""
        version = "Not determined"
        version_message = subprocess.run([exe,'-v'], capture_output=True)
        for line in version_message.stdout.decode().splitlines():
            if line.startswith('This'):
                info = line.split()
                if len(info) >= 5:
                    version = info[4]
                break
        return version

    def generate_partition_files(self, command: list, basename: str) -> list:
        """Generate the partition files enumerating invariant site counts"""
        current_wd = os.getcwd()
        if self.invariant_site_correction:
            partitions_fn = 'invariant_sites.' + os.path.basename(basename) + '.partitions'
            partition_fn = 'invariant_sites.' + os.path.basename(basename) + '.partition'
            with open(partitions_fn,'w') as partitions_file:
                partitions_file.write('[asc~' + partition_fn + '], ASC_DNA, p1=1-' + str(self.partition_length) + '\n')
                partitions_file.flush()
            with open(partition_fn,'w') as partition_file:
                partition_file.write(' '.join([str(x) for x in self.constant_base_counts]) + '\n')
                partition_file.flush()
            command.extend(["-q", partitions_fn])
        return command
        
    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the RAxML executable for tree building"""
        command = self.base_command.copy()
        if self.invariant_site_correction:
            # Write partition input files - https://cme.h-its.org/exelixis/resource/download/NewManual.pdf
            command = self.generate_partition_files(command,basename)
            # Complete command string
            command.extend(["-f", "d", "-p", str(1)])
        command.extend(["-s", alignment_filename, "-n", basename])
        if input_tree:
            command.extend(["-t", input_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def internal_sequence_reconstruction_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the RAxML executable for ancestral sequence reconstruction"""
        command = self.base_command.copy()
        command = self.generate_partition_files(command,basename)
        command.extend(["-f", "A", "-p", str(1)])
        command.extend(["-s", alignment_filename, "-n", basename])
        command.extend(["-t", input_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def select_executable_based_on_threads(self):
        """Chooses an appropriate executable"""
        if self.threads == 1:
            single_threaded_exec = utils.choose_executable_based_on_processor(
                self.single_threaded_executables)
            if single_threaded_exec is not None:
                return single_threaded_exec
            else:
                print("Trying multithreaded version of RAxML because no single threaded version of RAxML could be "
                      "found. Just to warn you, this requires 2 threads.\n")
                self.threads = 2

        if self.threads > 1:
            multi_threaded_exec = utils.choose_executable_based_on_processor(
                self.multi_threaded_executables)
            if multi_threaded_exec is not None:
                return multi_threaded_exec
            else:
                sys.stderr.write("No suitable RAxML version could be identified. Please try reinstalling the software\n")
                sys.exit(1)

    def convert_raw_ancestral_states_to_fasta(self, input_filename, output_filename):
        """Converts the file containing ancestral sequences into FASTA format"""
        with open(input_filename, 'r') as infile:
            with open(output_filename, 'w+') as outfile:
                for sequence_line in infile:
                    [sequence_name, sequence_bases] = sequence_line.split(' ')
                    sequence_bases = sequence_bases.replace('?', 'N')
                    outfile.write('>' + self.replace_internal_node_label(sequence_name) + '\n')
                    outfile.write(sequence_bases)

    def replace_internal_node_label(self, label):
        """Changes the label of internal nodes"""
        return self.internal_node_prefix + label

    def get_info_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the file containing the fitted model parameters"""
        fn = tmp + '/RAxML_info.' + basename + '_reconstruction'
        return fn
    
    def get_recontree_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the tree generated by model fitting"""
        fn = tmp + '/RAxML_result.' + basename + '_reconstruction'
        return fn

    def model_fitting_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Fits a nucleotide substitution model to a tree and an alignment"""
        command = self.base_command.copy()
        command = self.generate_partition_files(command,basename)
        command.extend(["-s", alignment_filename, "-n", os.path.basename(basename) + '_reconstruction', "-t", input_tree])
        command.extend(["-f e"])
        command.extend(["-w",os.path.dirname(basename)])
        return " ".join(command)
    
    def bootstrapping_command(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a bootstrapping analysis and annotates the nodes of a summary tree"""
        # Run bootstraps
        command = self.base_command.copy()
        command = self.generate_partition_files(command,basename)
        command.extend(["-s", alignment_filename, "-n", basename + ".bootstrapped_trees"])
        command.extend(["-w",tmp])
        command.extend(["-x",self.seed])
        command.extend(["-#",str(self.bootstrap)])
        # Output
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        command.extend([";"])
        return " ".join(command)

    def sh_test(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a single branch support test"""
        command = self.base_command.copy()
        command = self.generate_partition_files(command,basename)
        command.extend(["-f", "J"])
        command.extend(["-s", alignment_filename, "-n", input_tree + ".sh_support"])
        command.extend(["-t", input_tree])
        command.extend(["-w",tmp])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def get_bootstrapped_trees_file(self, tmp: str, basename: str) -> str:
        """Return bootstrapped tree files name"""
        file_name = tmp + "/RAxML_bootstrap." + basename + ".bootstrapped_trees"
        return file_name

class RAxMLNG:
    """Class for operations with the RAxML executable"""

    def __init__(self, threads: 1, model: str, constant_base_counts = [0,0,0,0], bootstrap = 0, seed = None, internal_node_prefix = "", verbose = False, invariant_site_correction = True, additional_args = None):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.model = model
        self.tree_prefix = ""
        self.tree_suffix = ".raxml.bestTree"
        self.asr_prefix = ""
        self.asr_suffix = ".raxml.ancestralStates"
        self.asr_tree_prefix = ""
        self.asr_tree_suffix = ".raxml.ancestralTree"
        self.alignment_suffix = ".phylip"
        self.internal_node_prefix = internal_node_prefix
        self.bootstrap = bootstrap
        self.seed = utils.set_seed(seed)
        self.constant_base_counts = constant_base_counts
        self.additional_args = additional_args
        self.invariant_site_correction = (invariant_site_correction if max(constant_base_counts) > 0 else False)

        self.single_threaded_executables = ['raxml-ng']
        self.multi_threaded_executables = ['raxml-ng']
        self.executable = self.select_executable_based_on_threads()
        if self.executable is None:
            sys.exit("No usable version of RAxML-NG could be found.")
        command = [self.executable]
        
        # Reproducibility
        self.name = 'RAxMLNG'
        self.version = self.get_version(self.executable)
        self.citation = "https://doi.org/10.1093/bioinformatics/btz305"
        
        # Set parallelisation
        command.extend(["--threads", str(self.threads)])

        # Add model
        command.extend(["--model"])
        invariant_site_string = ""
        if invariant_site_correction:
            invariant_site_string = "+ASC_STAM{" + '/'.join([str(x) for x in constant_base_counts]) + "}"
        if self.model == 'JC':
            command.extend(["JC"+invariant_site_string])
        elif self.model == 'K2P':
            command.extend(["K80"+invariant_site_string])
        elif self.model == 'HKY':
            command.extend(["HKY"+invariant_site_string])
        elif self.model == 'GTR':
            command.extend(["GTR"+invariant_site_string])
        elif self.model == 'GTRGAMMA':
            command.extend(["GTR+G"+invariant_site_string])
        else:
            command.extend([self.model])
        command.extend(["--seed",self.seed])
        # Additional arguments
        if self.additional_args is not None:
            command.extend([self.additional_args])
        self.base_command = command

    def get_version(self,exe) -> str:
        """Gets the version of the tree building algorithm being used"""
        version = "Not determined"
        version_message = subprocess.run([exe,'-v'], capture_output=True)
        for line in version_message.stdout.decode().splitlines():
            if line.startswith('RAxML-NG'):
                info = line.split()
                if len(info) >= 3:
                    version = info[2]
                break
        return version

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the RAxMLNG executable for tree building"""
        command = self.base_command.copy()
        if "--search1" not in command:
            command.extend(["--search"])
        command.extend(["--msa", alignment_filename, "--prefix", basename])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def internal_sequence_reconstruction_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the RAxMLNG executable for ancestral sequence reconstruction"""
        command = self.base_command.copy()
        command.extend(["--ancestral"])
        command.extend(["--msa", alignment_filename, "--prefix", basename])
        command.extend(["--tree", input_tree])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def select_executable_based_on_threads(self):
        """Chooses an appropriate executable"""
        if self.threads == 1:
            single_threaded_exec = utils.choose_executable_based_on_processor(
                self.single_threaded_executables)
            if single_threaded_exec is not None:
                return single_threaded_exec
            else:
                print("Trying multithreaded version because no single threaded version could be "
                      "found. Just to warn you, this requires 2 threads.\n")
                self.threads = 2

        if self.threads > 1:
            multi_threaded_exec = utils.choose_executable_based_on_processor(
                self.multi_threaded_executables)
            if multi_threaded_exec is not None:
                return multi_threaded_exec
            else:
                return None

    def convert_raw_ancestral_states_to_fasta(self, input_filename, output_filename):
        """Converts the file containing ancestral sequences into FASTA format"""
        # Use the IUPAC codes from here: https://www.bioinformatics.org/sms/iupac.html
        with open(input_filename, 'r') as infile:
            with open(output_filename, 'w+') as outfile:
                for sequence_line in infile:
                    [sequence_name, sequence_bases] = sequence_line.split()
                    sequence_bases = sequence_bases.translate(str.maketrans({'?': 'N',
                                                                             'R': 'N',
                                                                             'Y': 'N',
                                                                             'S': 'N',
                                                                             'W': 'N',
                                                                             'K': 'N',
                                                                             'M': 'N',
                                                                             'B': 'N',
                                                                             'D': 'N',
                                                                             'H': 'N',
                                                                             'V': 'N'
                                                                            }
                                                                     )
                                                              )
                    outfile.write('>' + self.replace_internal_node_label(sequence_name) + '\n')
                    outfile.write(sequence_bases + '\n')

    def replace_internal_node_label(self, label):
        """Changes the label of internal nodes"""
        return self.internal_node_prefix + label

    def get_info_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the file containing the fitted model parameters"""
        fn = tmp + '/' + basename + '_reconstruction.raxml.bestModel'
        return fn
    
    def get_recontree_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the tree generated by model fitting"""
        fn = tmp + '/' + basename + '_reconstruction.raxml.bestTree'
        return fn

    def model_fitting_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Fits a nucleotide substitution model to a tree and an alignment"""
        command = self.base_command.copy()
        command.extend(["--evaluate"])
        command.extend(["--msa", alignment_filename, "--prefix", basename + '_reconstruction', "--tree", input_tree])
        return " ".join(command)

    def bootstrapping_command(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a bootstrapping analysis and annotates the nodes of a summary tree"""
        # Run bootstraps
        command = self.base_command.copy()
        command.extend(["--bootstrap"])
        command.extend(["--msa", alignment_filename, "--prefix", tmp + "/" + basename])
        command.extend(["--bs-trees",str(self.bootstrap)])
        # Output
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        command.extend([";"])
        return " ".join(command)

    def annotate_tree_using_bootstraps_command(self, alignment_filename: str, input_tree: str, bootstrapped_trees: str, basename: str, tmp: str, transfer = False) -> str:
        # Annotate tree with bootstraps
        command = self.base_command.copy()
        command.extend(["--support"])
        command.extend(["--bs-trees",bootstrapped_trees])
        command.extend(["--tree",input_tree])
        command.extend(["--prefix",tmp + "/" + basename + ".bootstrapped"])
        if transfer:
            command.extend(["--bs-metric tbe"])
        else:
            command.extend(["--bs-metric fbp"])
        # Output
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        command.extend([";"])
        # Rename final file
        command.extend(["cp",tmp + "/" + basename + ".bootstrapped.raxml.support", basename + ".tre.bootstrapped"])
        return " ".join(command)

    def get_bootstrapped_trees_file(self, tmp: str, basename: str) -> str:
        """Return bootstrapped tree files name"""
        file_name = tmp + "/" + basename + ".raxml.bootstraps"
        return file_name

class VeryFastTree:
    """Class for operations with the VeryFastTree executable"""

    def __init__(self, threads: int, bootstrap = 0, model='GTRCAT', seed = None, verbose = False, additional_args = None):
        """Initialises the object"""
        self.verbose = verbose
        self.threads = threads
        self.model = model
        self.tree_prefix = ""
        self.tree_suffix = ".tre"
        self.alignment_suffix = ".snp_sites.aln"
        self.bootstrap = bootstrap
        self.additional_args = additional_args
        self.seed = utils.set_seed(seed)
        self.invariant_site_correction = False

        # Identify executable
        self.potential_executables = ["VeryFastTree", "veryfasttree"]
        self.executable = utils.choose_executable(self.potential_executables)
        if self.executable is None:
            sys.exit("No usable version of VeryFastTree could be found.")

        # Reproducibility
        self.name = 'VeryFastTree'
        self.version = self.get_version(self.executable)
        self.citation = "https://doi.org/10.1093/gigascience/giae055"

        # Function for returning base command
        command = [self.executable]
        command.extend(["-nt"])
        command.extend(["-rawdist"]) # https://morgannprice.github.io/fasttree/
        if self.model == 'JC':
            command.extend(["-nocat"])
        elif self.model == 'GTR':
            command.extend(["-gtr","-nocat"])
        elif self.model == 'GTRGAMMA':
            command.extend(["-gtr","-gamma"])
        elif self.model == 'GTRCAT':
            command.extend(["-gtr"])
        else:
            command.extend([self.model])
        command.extend(["-seed",self.seed])
        # Set number threads
        command.extend(["-threads",str(self.threads)])
        # Additional arguments
        if self.additional_args is not None:
            command.extend([self.additional_args])
        # Define final command
        self.base_command = command

    def get_version(self,exe) -> str:
        """Gets the version of the tree building algorithm being used"""
        version = "Not determined"
        version_message = subprocess.run([exe], capture_output=True)
        for line in version_message.stderr.decode().splitlines():
            if line.startswith('VeryFastTree'):
                info = line.split()
                version = info[1]
                break
        return version

    def tree_building_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Constructs the command to call the VeryFastTree executable"""
        command = self.base_command.copy()
        # N.B. version 4.0.4 - intree appears to cause a problem so is not used in command construction
#        if input_tree:
#            command.extend(["-intree", input_tree])
        output_tree = basename + self.tree_suffix
        command.extend(["-nosupport"])
        command.extend(["-out", output_tree])
        command.extend(["-log", basename + '.log'])
        command.append(alignment_filename)
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)
    
    def get_info_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the file containing the fitted model parameters"""
        fn = tmp + '/' + basename + '.log'
        return fn
    
    def get_recontree_filename(self, tmp: str, basename: str) -> str:
        """Returns the name of the tree generated by model fitting"""
        fn = tmp + '/' + basename + '.treefile'
        return fn

    def model_fitting_command(self, alignment_filename: str, input_tree: str, basename: str) -> str:
        """Fits a nucleotide substitution model to a tree and an alignment"""
        command = self.base_command.copy()
        command.extend(["-mllen","-nome"])
        command.extend(["-nosupport"])
        command.extend(["-intree",input_tree])
        command.extend(["-log", basename + ".log"])
        command.extend(["-out", basename + ".treefile"])
        command.extend([alignment_filename])
        return " ".join(command)
        
    def bootstrapping_command(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a bootstrapping analysis and annotates the nodes of a summary tree"""
        command = self.base_command.copy()
        output_tree = basename + self.tree_suffix
        command.extend(["-nosupport"])
        command.extend(["-intree1",input_tree]) # http://www.microbesonline.org/fasttree/
        command.extend(["-out", tmp + "/" + basename + ".bootstrapped_trees"])
        command.extend(["-log", basename + ".log"])
        command.extend(["-n", str(self.bootstrap)])
        command.append(alignment_filename + ".bootstrapping.aln")
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)
    
    def sh_test(self, alignment_filename: str, input_tree: str, basename: str, tmp: str) -> str:
        """Runs a single branch support test"""
        command = self.base_command.copy()
        command.extend(["-mllen","-nome"])
        command.extend(["-intree",input_tree])
        command.extend(["-out",tmp + "/" + input_tree + ".sh_support"])
        command.extend([alignment_filename])
        if not self.verbose:
            command.extend([">", "/dev/null", "2>&1"])
        return " ".join(command)

    def get_bootstrapped_trees_file(self, tmp: str, basename: str) -> str:
        """Return bootstrapped tree files name"""
        file_name = tmp + "/" + basename + ".bootstrapped_trees"
        return file_name



#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020 Imperial College London and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import argparse
import re
from collections import Counter

def parse_input_args():

    parser = argparse.ArgumentParser(
        description='Script to evaluate and reformat an alignment prior to Gubbins analysis',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--aln',
                        '-a',
                        dest="aln",
                        help='Multifasta alignment filename',
                        required=True)
    parser.add_argument('--out-aln',
                        dest="out_aln",
                        help='Reformatted alignment filename',
                        default = None,
                        required=False)
    parser.add_argument('--out',
                        '-o',
                        dest="out",
                        help='Output CSV filename',
                        required=True)

    return parser.parse_args()

def main(input_args):

    # Read the number of rows in the alignment
    row_num = 0
    tot_lines = 0
    with open(input_args.aln, "r") as aln_file:
        for line in aln_file:
            if re.search("^>", line):
                row_num += 1
            tot_lines += 1

    # Filter alignment if requested
    aln_filename = input_args.aln
    if input_args.out_aln:
        with open(input_args.aln, "r") as aln_file, \
                open(input_args.out_aln, "w") as out_aln_file:
            for line in aln_file:
                if re.search("^>", line):
                    name = line.replace("#","_").replace(":","_").replace(">","").rstrip().split()
                    out_aln_file.write('>' + name[0] + '\n')
                else:
                    sequence = line.upper().rstrip()
                    sequence = re.sub('[^ACGTN-]','N',sequence)
                    out_aln_file.write(sequence + '\n')
        aln_filename = input_args.out_aln

    # Going to use counter in a first pass to store sequence counts and the range of sequences
    total_base_counts = []
    iso_data = []
    total_headers = []
    tot_length_str = str(tot_lines)
    print("Running through alignment file: %s" % input_args.aln)
    print()
    with open(aln_filename, "r") as aln_file:
        for index,line in enumerate(aln_file):
            num_zeros = len(tot_length_str) - len(str(index + 1))
            fmt_index = (("0" * num_zeros) + str(index + 1))
            
            if re.search("^>", line):
                iso = re.sub("\n","",(re.sub("^>","",line)))
                iso_data.append(iso)
                
            else:
                current_count = Counter(line.strip())
                iso_headers = sorted(list(current_count))
                add_list = list(set(iso_headers) - set(total_headers))
                if len(add_list) > 0:
                    total_headers.extend(add_list)
                total_base_counts.append(current_count)
            print("Counted %s of %s rows" % (fmt_index, tot_length_str), end="\r", flush=True)

    ## Second pass to line up all the counts across the isolates 
    print("")
    print("Assessing counts...")
    total_headers.sort()
    isolate_bases = []
    for count in total_base_counts:
        iso_row = []
        for header in total_headers:
            iso_row.append(count[header])
        isolate_bases.append(iso_row)

    total_headers.insert(0, "isolate")
    write_out_headers = [re.sub("-","gap",i) for i in total_headers]
    print("Writing out results...")
    with open((input_args.out), "w") as output:
        output.write(",".join(write_out_headers) + "\n")
        for i, aln_row in enumerate(isolate_bases):
            aln_row_str = list(map(str, aln_row))
            aln_row_str.insert(0, iso_data[i])
            output.write(",".join(aln_row_str) + "\n")

    print("Finished")

if __name__ == '__main__':
    input_args = parse_input_args()
    main(input_args)





#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020  Wellcome Trust Sanger Institute and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import sys
import argparse
import re
# Biopython imports
from Bio import AlignIO
from Bio import Phylo
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.Seq import MutableSeq

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Mask recombinant regions detected by'
                                                    ' Gubbins from the input alignment',
                                     prog='mask_gubbins_aln')

    # input options
    parser.add_argument('--aln',
                        help = 'Input alignment (FASTA format)',
                        required = True)
    parser.add_argument('--gff',
                        help = 'GFF of recombinant regions detected by Gubbins',
                        required = True)
    parser.add_argument('--out',
                        help = 'Output file name',
                        required = True)
    parser.add_argument('--out-fmt',
                        help = 'Format of output alignment',
                        default = 'fasta')
    parser.add_argument('--missing-char',
                        help = 'Character used to replace recombinant sequence',
                        default = '-')

    return parser.parse_args()

# main code
if __name__ == "__main__":

    # Get command line options
    args = get_options()
    
    # Read in alignment
    overall_taxon_list = []
    alignment = AlignIO.read(args.aln,'fasta')
    for taxon in alignment:
        overall_taxon_list.append(taxon.id)
        taxon.seq = MutableSeq(taxon.seq)
    overall_taxon_set = set(overall_taxon_list)
    
    # Read recombinant regions from GFF
    taxon_pattern = re.compile('taxa="([^"]*)"')
    with open(args.gff,'r') as gff_file:
        for line in gff_file.readlines():
            if not line.startswith('##'):
                info = line.rstrip().split('\t')
                start = int(info[3])-1
                end = int(info[4])-1
                taxon_set = set(taxon_pattern.search(info[8]).group(1).split())
                if taxon_set.issubset(overall_taxon_set):
                    for taxon in alignment:
                        if taxon.id in taxon_set:
                            taxon.seq[start:end+1] = args.missing_char*(end-start+1)

    # Write out masked sequence
    with open(args.out,'w') as out_aln:
        AlignIO.write(alignment, out_aln, args.out_fmt)



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests the construction of file names and commands with no dependencies on external applications.
"""

import unittest.mock
import os
import filecmp
from gubbins import treebuilders, utils

modules_dir = os.path.dirname(os.path.abspath(treebuilders.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestStringConstruction(unittest.TestCase):

    def test_iqtree_convert_raw_ancestral_states_to_fasta(self):
        iqtree = treebuilders.IQTree(1, model = 'GTRGAMMA')
        iqtree.convert_raw_ancestral_states_to_fasta(os.path.join(data_dir, 'iqtree_ancestral.state'),
                                                     os.path.join(data_dir, 'iqtree_ancestral.fasta.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'iqtree_ancestral.fasta.actual'),
                           os.path.join(data_dir, 'iqtree_ancestral.fasta.expected'))
        os.remove(os.path.join(data_dir, 'iqtree_ancestral.fasta.actual'))

    def test_iqtree_replace_internal_node_label(self):
        iqtree = treebuilders.IQTree(1, model = 'GTRGAMMA', internal_node_prefix='AAA')
        assert iqtree.replace_internal_node_label('Node20') == 'AAA20'

    def test_raxml_select_executable_based_on_threads(self):

        raxml_st = treebuilders.RAxML(1, model = 'GTRGAMMA')
        raxml_mt = treebuilders.RAxML(8, model = 'GTRGAMMA')
        assert (raxml_st.select_executable_based_on_threads() != raxml_mt.select_executable_based_on_threads() or raxml_st.threads > 1)

    def test_raxml_convert_raw_ancestral_states_to_fasta(self):
        raxml = treebuilders.RAxML(8)
        raxml.convert_raw_ancestral_states_to_fasta(os.path.join(data_dir, 'raxml_ancestral.state'),
                                                    os.path.join(data_dir, 'raxml_ancestral.fasta.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'raxml_ancestral.fasta.actual'),
                           os.path.join(data_dir, 'raxml_ancestral.fasta.expected'))
        os.remove(os.path.join(data_dir, 'raxml_ancestral.fasta.actual'))

    def test_raxml_replace_internal_node_label(self):
        raxml = treebuilders.RAxML(1, internal_node_prefix='AAA')
        assert raxml.replace_internal_node_label('Node20') == 'AAANode20'


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import setuptools

setuptools.setup(
    name='gubbins',
    version = open('VERSION').read().strip(),
    description='Frontend to the Gubbins BioInformatics tool',
    author='Andrew Page, Nicholas Croucher, Aidan Delaney, Christoph Puethe and Simon Harris',
    author_email='n.croucher@imperial.ac.uk',
    url='https://github.com/nickjcroucher/gubbins/',
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "run_gubbins.py = gubbins.run_gubbins:main",
        ]
    },
    scripts=[
        'scripts/generate_ska_alignment.py',
        'scripts/extract_gubbins_clade.py',
        'scripts/mask_gubbins_aln.py',
        'scripts/gubbins_alignment_checker.py',
        'scripts/generate_files_for_clade_analysis.py',
        'scripts/count_recombinations_per_gene.py',
        'scripts/extract_recombinant_sequences.py',
        'scripts/extract_gubbins_clade_statistics.py',
        '../R/scripts/plot_gubbins.R'
    ],
    tests_require=[
        "pytest >= 4.6",
        "wheel >= 0.34",
        "biopython >= 1.59",
        "dendropy  >= 4.0.2",
        "multiprocess >= 0.70",
        "scipy >= 1.5.3",
        "numpy >= 1.19",
        "ska >= 1.0",
        "numba > 0.0"
    ],
    long_description="""\
      Gubbins is a tool that generates a reconstruction of
      a microbial strain's recent evolutionary history through
      identifying imports of divergent sequence through recombination
      and generating a phylogeny from the remaining clonal frame.
      """,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=[
        "biopython >= 1.59",
        "dendropy  >= 4.0.2",
        "multiprocess >= 0.70",
        "scipy >= 1.5.3",
        "numpy >= 1.19",
        "numba > 0.0"
    ],
    license="GPL"
)



#!/usr/bin/env python
# encoding: utf-8

import sys
import hashlib
from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from collections import defaultdict


class PreProcessFasta(object):

    def __init__(self, input_filename, verbose=False, filter_percentage=25):
        self.input_filename = input_filename
        self.verbose = verbose
        self.filter_percentage = filter_percentage

    def hash_sequences(self):
        sequence_hash_to_taxa = defaultdict(list)
        with open(self.input_filename) as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            for alignment in alignments:
                for record in alignment:
                    sequence_hash = hashlib.md5()
                    sequence_hash.update(str(record.seq).encode('utf-8'))
                    hash_of_sequence = sequence_hash.digest()
                    sequence_hash_to_taxa[hash_of_sequence].append(record.id)

                    if self.verbose:
                        print("Sample " + str(record.id) + " has a hash of " + str(hash_of_sequence))
        input_handle.close()
        return sequence_hash_to_taxa

    def calculate_sequences_missing_data_percentage(self):
        sequences_to_missing_data = {}
        with open(self.input_filename) as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            for alignment in alignments:
                for record in alignment:
                    number_of_gaps = 0
                    number_of_gaps += record.seq.count('n')
                    number_of_gaps += record.seq.count('N')
                    number_of_gaps += record.seq.count('-')
                    sequence_length = len(record.seq)

                    if sequence_length == 0:
                        sequences_to_missing_data[record.id] = 100
                        if self.verbose:
                            print("Sample " + str(record.id) + " has no sequence ")
                    else:
                        per_missing_data = number_of_gaps*100/sequence_length
                        sequences_to_missing_data[record.id] = per_missing_data
                        if self.verbose:
                            print("Sample " + str(record.id) + " has missing data percentage of " +
                                  str(per_missing_data))

        input_handle.close()
        return sequences_to_missing_data

    def taxa_missing_too_much_data(self):
        taxa_to_remove = []
        for taxa, percentage_missing in self.calculate_sequences_missing_data_percentage().items():
            if percentage_missing > self.filter_percentage:
                taxa_to_remove.append(taxa)
                print("Excluded sequence " + taxa + " because it had " + str(percentage_missing) +
                      " percentage missing data while a maximum of " + str(self.filter_percentage) + " is allowed")

        return taxa_to_remove

    def taxa_of_duplicate_sequences(self):
        taxa_to_remove = []
        for sequence_hash, taxa in sorted(self.hash_sequences().items()):
            if len(taxa) > 1:
                taxon_to_keep = taxa.pop()
                for taxon in taxa:
                    print("Sequences in " + taxon + " and " + taxon_to_keep + " are identical, removing " + taxon +
                          " from analysis")
                    taxa_to_remove.append(taxon)

        return taxa_to_remove

    def remove_duplicate_sequences_and_sequences_missing_too_much_data(self, output_filename,
                                                                       remove_identical_sequences=None):

        if not remove_identical_sequences:
            taxa_to_remove = self.taxa_missing_too_much_data()
        else:
            taxa_to_remove = self.taxa_of_duplicate_sequences() + self.taxa_missing_too_much_data()

        with open(self.input_filename) as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            output_alignments = []
            number_of_included_alignments = 0
            for alignment in alignments:
                for record in alignment:
                    if record.id not in taxa_to_remove:
                        output_alignments.append(record)
                        number_of_included_alignments += 1
            if number_of_included_alignments <= 1:
                sys.exit("Not enough sequences are left after removing duplicates.Please check you input data.")

        with open(output_filename, "w") as output_handle:
            AlignIO.write(MultipleSeqAlignment(output_alignments), output_handle, "fasta")

        return taxa_to_remove

    def get_sequence_names(self):
        sequence_names = []
        with open(self.input_filename) as input_handle:
            alignments = AlignIO.parse(input_handle, "fasta")
            for alignment in alignments:
                for record in alignment:
                    sequence_names.append(record.id)
        return sequence_names
    
    def get_alignment_length(self, format = "fasta"):
        alignment = AlignIO.read(self.input_filename, format)
        return alignment.get_alignment_length()

    def get_alignment_information(self, format = "fasta"):
        sequence_names = []
        base_frequencies = [0,0,0,0]
        alignment = AlignIO.read(self.input_filename, format)
        alignment_length = alignment.get_alignment_length()
        
        for record in alignment:
            sequence_names.append(record.id)
            for index,base in enumerate(['A','C','G','T']):
                base_frequencies[index] += (record.seq.count(base) + record.seq.count(base.lower()))
        
        base_frequencies = [b/(alignment_length*len(sequence_names)) for b in base_frequencies]
        
        return alignment_length, sequence_names, base_frequencies



#! /usr/bin/env python3
# encoding: utf-8

"""
Testing the python scripts work
"""

import unittest
import os
import shutil
import subprocess
import hashlib
import glob
from gubbins import common, run_gubbins

modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
preprocess_dir = os.path.join(data_dir, 'preprocessfasta')
working_dir = os.path.join(modules_dir, 'tests')

class TestPythonScripts(unittest.TestCase):

    ## Test the alignment_checker script
    def test_alignment_checker(self):
        small_aln = os.path.join(data_dir, "valid_alignment.aln")
        output_csv = os.path.join(working_dir, "valid_alignment_test.csv")
        test_csv = os.path.join(data_dir, "test_valid_output.csv")
        aln_cmd = "gubbins_alignment_checker.py --aln " + small_aln + " --out " + output_csv
        subprocess.check_call(aln_cmd, shell=True)
        assert self.md5_check(output_csv, test_csv)
        os.remove(output_csv)

    def test_alignment_reformatting(self):
        invalid_aln = os.path.join(data_dir, "invalid_alignment.aln")
        output_file = os.path.join(working_dir, "invalid_alignment_test.aln")
        output_csv = os.path.join(working_dir, "valid_alignment_test.csv")
        test_aln = os.path.join(data_dir, "corrected_alignment.aln")
        test_csv = os.path.join(data_dir, "corrected_alignment.csv")
        aln_cmd = "gubbins_alignment_checker.py --aln " + invalid_aln + " --out-aln " + output_file + " --out " + output_csv
        subprocess.check_call(aln_cmd, shell=True)
        assert self.md5_check(output_csv, test_csv)
        assert self.md5_check(output_file, test_aln)
        os.remove(output_csv)
        os.remove(output_file)

    ## Test the clade extraction script
    def test_clade_extraction(self):
        multiple_aln = os.path.join(data_dir, "multiple_recombinations.aln")
        clade_list = os.path.join(data_dir, "clade_to_extract.txt")
        multiple_gff = os.path.join(data_dir, "multiple_recombinations_gubbins.recombination_predictions.gff")
        multiple_tree = os.path.join(data_dir, "multiple_recombinations_gubbins.node_labelled.final_tree.tre")
        out_aln = os.path.join(data_dir, "multiple_recombinations_extract.aln")
        out_gff = os.path.join(data_dir, "multiple_recombinations_extract.gff")
        out_tree = os.path.join(data_dir, "multiple_recombinations_extract.tree")
        base_path = os.path.join(data_dir, "multiple_recombinations_extract")
        ## Get the test files 
        test_aln = os.path.join(data_dir, "multiple_recombinations_clade_extract.aln")
        test_gff = os.path.join(data_dir, "multiple_recombinations_clade_extract.gff")
        test_tree = os.path.join(data_dir, "multiple_recombinations_clade_extract.tree")
        # Script name
        extract_clade_cmd = "extract_gubbins_clade.py --list " + clade_list + " --aln " + multiple_aln +\
            " --gff " + multiple_gff + " --tree " + multiple_tree + " --out " + base_path +  " --out-fmt fasta"
        subprocess.check_call(extract_clade_cmd, shell=True)
        assert self.md5_check(out_aln, test_aln)
        assert self.md5_check(out_gff, test_gff)
        assert self.md5_check(out_tree, test_tree)
        os.remove(out_aln)
        os.remove(out_gff)
        os.remove(out_tree)

    ## Test the masking aln script
    def test_masking_aln(self):
        multiple_aln = os.path.join(data_dir, "multiple_recombinations.aln")
        multiple_gff = os.path.join(data_dir, "multiple_recombinations_gubbins.recombination_predictions.gff")
        out_aln = os.path.join(data_dir, "multiple_recombinations_mask.aln")
        ## Get the test file 
        test_aln = os.path.join(data_dir, "masking_multiple.aln")
        # Script name
        extract_clade_cmd = "mask_gubbins_aln.py --aln " + multiple_aln +\
            " --gff " + multiple_gff + " --out " + out_aln +  " --out-fmt fasta"
        subprocess.check_call(extract_clade_cmd, shell=True)
        assert self.md5_check(out_aln, test_aln)
        os.remove(out_aln)

    ## Test the ska alignment generator 
    def test_generate_ska_alignment(self):
        exit_code = 1
        ## Run the generate_ska_alignment script
        fasta_loc = os.path.join(preprocess_dir, './ska_fasta_list.txt')
        with open(fasta_loc,'w') as list_file:
            for i in range(1,7):
                list_file.write('sequence_t' + str(i) + '\t' + \
                    os.path.join(preprocess_dir,'sequence_t' + str(i) + '.fasta\n'))
        ref_seq = os.path.join(preprocess_dir, 'sequence_t1.fasta')
        aln_out = os.path.join(preprocess_dir, 'ska_test_aln.aln')
        # Script name
        ska_cmd = "generate_ska_alignment.py --input " + fasta_loc +\
            " --reference " + ref_seq + " --out " + aln_out +\
                " --k 13"
        subprocess.check_call(ska_cmd, shell=True)
        ## Now run gubbins on the aln and check all the output is produced 
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--prefix", "ska_test",
                                                    "--verbose", "--mar",
                                                    aln_out]))
        exit_code = self.check_for_output_files('ska_test')
        self.cleanup('ska_test')
        os.remove(aln_out)
        os.remove(fasta_loc)
        assert exit_code == 0

    # Test clade file extraction script
    def test_clade_file_generation(self):
        multiple_aln = os.path.join(data_dir, "multiple_recombinations.aln")
        clade_list = os.path.join(data_dir, "clade_to_extract.txt")
        multiple_gff = os.path.join(data_dir, "multiple_recombinations_gubbins.recombination_predictions.gff")
        multiple_tree = os.path.join(data_dir, "multiple_recombinations_gubbins.node_labelled.final_tree.tre")
        out_aln = os.path.join(data_dir, "multiple_recombinations_generate.aln")
        out_gff = os.path.join(data_dir, "multiple_recombinations_generate.gff")
        out_tree = os.path.join(data_dir, "multiple_recombinations_generate.tree")
        base_path = os.path.join(data_dir, "multiple_recombinations_generate")
        ## Get the test files
        test_aln = os.path.join(data_dir, "multiple_recombinations_clade_generate.aln")
        test_gff = os.path.join(data_dir, "multiple_recombinations_clade_generate.gff")
        test_tree = os.path.join(data_dir, "multiple_recombinations_clade_generate.tree")
        # Script name
        clade_file_generation_cmd = "generate_files_for_clade_analysis.py --list " + clade_list + " --aln " + multiple_aln +\
            " --gff " + multiple_gff + " --tree " + multiple_tree + " --out " + base_path
        subprocess.check_call(clade_file_generation_cmd, shell=True)
        assert self.md5_check(out_aln, test_aln)
        assert self.md5_check(out_gff, test_gff)
        assert self.md5_check(out_tree, test_tree)
        os.remove(out_aln)
        os.remove(out_gff)
        os.remove(out_tree)

    # Test clade file extraction script
    def test_recombination_counting_per_gene(self):
        multiple_gff = os.path.join(data_dir, "multiple_recombinations_gubbins.recombination_predictions.gff")
        annnotation_gff = os.path.join(data_dir, "multiple_recombinations_annotation.gff")
        out_tab = os.path.join(data_dir, "test_multiple_recombinations_annotation.gff")
        check_tab = os.path.join(data_dir, "multiple_recombinations_rec_per_gene.tab")
        rec_count_cmd = "count_recombinations_per_gene.py --rec-gff " + multiple_gff + " --anno-gff " + annnotation_gff + " --out " + out_tab
        subprocess.check_call(rec_count_cmd, shell=True)
        assert self.md5_check(out_tab, check_tab)
        os.remove(out_tab)

    # Test clade file extraction script
    def test_extract_recombination_sequences(self):
        multiple_aln = os.path.join(data_dir, "multiple_recombinations.aln")
        multiple_gff = os.path.join(data_dir, "multiple_recombinations_gubbins.recombination_predictions.gff")
        out_dir = os.path.join(data_dir, "test_loci")
        rec_count_cmd = "extract_recombinant_sequences.py --aln " + multiple_aln + " --gff " + multiple_gff + " --out-dir " + out_dir
        subprocess.check_call(rec_count_cmd, shell=True)
        file_count = int(subprocess.run('ls -1 ' + out_dir + ' | wc -l', shell = True, stdout=subprocess.PIPE).stdout.decode('utf-8').rstrip())
        assert file_count > 1
        shutil.rmtree(out_dir)

    # Test plotting script (an R exception)
    def test_recombination_counting_per_gene(self):
        tree_input = os.path.join(data_dir, "expected_RAxML_result.multiple_recombinations.iteration_5")
        rec_input = os.path.join(data_dir, "multiple_recombinations_gubbins.recombination_predictions.gff")
        anno_input = os.path.join(data_dir, "test_annotation.gff")
        markup_input = os.path.join(data_dir, "test_markup.csv")
        meta_input = os.path.join(data_dir, "test_epi.csv")
        clades_input = os.path.join(data_dir, "test_clades.csv")
        fig_output = os.path.join(data_dir, "test_plot.pdf")
        plot_cmd = "plot_gubbins.R --tree " + tree_input + " --rec " + rec_input + " --output " + fig_output
        subprocess.check_call(plot_cmd, shell=True)
        plot_cmd = plot_cmd + " --annotation " +  anno_input
        subprocess.check_call(plot_cmd, shell=True)
        plot_cmd = plot_cmd + " --clades " + clades_input
        subprocess.check_call(plot_cmd, shell=True)
        plot_cmd = plot_cmd + " --markup " + markup_input
        subprocess.check_call(plot_cmd, shell=True)
        plot_cmd = plot_cmd + " --meta " + meta_input
        subprocess.check_call(plot_cmd, shell=True)
        os.remove(fig_output)

    @staticmethod
    def check_for_output_files(prefix):
        assert os.path.exists(prefix + '.summary_of_snp_distribution.vcf')
        assert os.path.exists(prefix + '.recombination_predictions.embl')
        assert os.path.exists(prefix + '.per_branch_statistics.csv')
        assert os.path.exists(prefix + '.filtered_polymorphic_sites.fasta')
        assert os.path.exists(prefix + '.filtered_polymorphic_sites.phylip')
        assert os.path.exists(prefix + '.recombination_predictions.gff')
        assert os.path.exists(prefix + '.branch_base_reconstruction.embl')
        assert os.path.exists(prefix + '.final_tree.tre')
        assert os.path.exists(prefix + '.node_labelled.final_tree.tre')
        return 0

    @staticmethod
    def cleanup(prefix):
        #os.chdir(working_dir)
        regex_to_remove = prefix + ".*"
        for file in glob.glob(regex_to_remove):
            os.remove(file)
        tmp_to_remove = "./tmp*/*"
        for file in glob.glob(tmp_to_remove):
            os.remove(file)
        for dir in glob.glob("./tmp*"):
            if os.path.isdir(dir):
                os.rmdir(dir)

    
    @staticmethod
    def md5_check(file_path, correct_output):
        file_sum = hashlib.md5(open(file_path, "rb").read()).hexdigest()
        correct_sum = hashlib.md5(open(correct_output, "rb").read()).hexdigest()
        return file_sum == correct_sum

if __name__ == "__main__":
    unittest.main(buffer=True)




#!/usr/bin/env python
# encoding: utf-8

import os
import re
import sys
from Bio import AlignIO
from collections import Counter
from Bio.Align import MultipleSeqAlignment
# Gubbins imports
from gubbins.utils import process_sequence_names

class ValidateFastaAlignment(object):

    def __init__(self, input_filename):
      self.input_filename = input_filename

    def is_input_fasta_file_valid(self):
      try:
          if not self.does_each_sequence_have_the_same_length():
              print("Each sequence must be the same length")
              return False
          if not self.are_sequence_names_unique():
              print("All sequence names in the fasta file must be unique")
              return False
          if not self.does_each_sequence_have_a_name_and_genomic_data():
              print("Each sequence must have a name and some genomic data")
              return False
      except:
          return False
      return True

    def does_each_sequence_have_a_name_and_genomic_data(self):
      with open(self.input_filename, "r") as input_handle:
        alignments = AlignIO.parse(input_handle, "fasta")
        number_of_sequences = 0
        for alignment in alignments:
            for record in alignment:
                number_of_sequences +=1
                if record.name is None or record.name == "":
                  sys.stderr.write("Error with the input FASTA file: " + record.name + " is blank\n")
                  return False
                if record.seq is None or record.seq == "":
                  sys.stderr.write("Error with the input FASTA file: " + record.name + " is empty\n")
                  return False
                if re.search('[^ACGTNacgtn-]', str(record.seq))  != None:
                  sys.stderr.write("Error with the input FASTA file: " + record.name + " contains disallowed characters, only ACGTNacgtn- are permitted\n")
                  return False
      return True

    def does_each_sequence_have_the_same_length(self):
      try:
        with open(self.input_filename) as input_handle:
          alignments = AlignIO.parse(input_handle, "fasta")
          sequence_length = -1
          for alignment in alignments:
              for record in alignment:
                 if sequence_length == -1:
                   sequence_length = len(record.seq)
                 elif sequence_length != len(record.seq):
                   print("Error with the input FASTA file: The sequences are not of the same length, this is not an alignment: "+record.name)
                   return False
      except:
        print("Unexpected error:", sys.exc_info()[0])
        print("Error with the input FASTA file: It is in the wrong format, check it is an alignment")
        return False
      return True

    def are_sequence_names_unique(self):
        any_modified_names = False
        with open(self.input_filename) as input_handle:
            alignment = AlignIO.read(input_handle, "fasta")
            sequence_names = []
            for record in alignment:
                # Remove disallowed characters
                if '#' in record.name or ':' in record.name:
                    record.name = process_sequence_names(record.name)
                    record.id = process_sequence_names(record.id)
                    record.description = process_sequence_names(record.description)
                    any_modified_names = True
                # Store modified names
                sequence_names.append(record.name)
        duplicate_sequence_list = [k for k,v in list(Counter(sequence_names).items()) if v>1]
        if duplicate_sequence_list != []:
            print("Duplicate sequences found after name processing:")
            for dup in duplicate_sequence_list:
                print(dup)
            return False
        # Update alignment if names changed
        if any_modified_names:
            with open(self.input_filename, "w") as output_handle:
                AlignIO.write(alignment,output_handle, "fasta")
        return True
      



#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020  Wellcome Trust Sanger Institute and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import sys
import argparse
import re
# Phylogenetic imports
import dendropy
# Biopython imports
from Bio import AlignIO
from Bio import Phylo
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Extract a clade from a Gubbins output',
                                     prog='extract_clade')

    # input options
    parser.add_argument('--list',
                        help = 'List of sequences to extract',
                        required = True)
    parser.add_argument('--aln',
                        help = 'Input alignment (FASTA format)',
                        required = True)
    parser.add_argument('--gff',
                        help = 'GFF of recombinant regions detected by Gubbins',
                        required = True)
    parser.add_argument('--tree',
                        help = 'Final tree generated by Gubbins',
                        required = True)
    parser.add_argument('--out',
                        help = 'Output file prefix',
                        required = True)
    parser.add_argument('--out-fmt',
                        help = 'Format of output alignment',
                        default = 'fasta')
    parser.add_argument('--missing-char',
                        help = 'Character used to replace recombinant sequence',
                        default = '-')

    return parser.parse_args()

# main code
if __name__ == "__main__":

    # Get command line options
    args = get_options()
    
    # Parse list of input sequences
    subset = set()
    # Read in FASTA assemblies
    with open(args.list,'r') as seq_list:
        for line in seq_list.readlines():
            subset.add(line.strip().split()[0])
    
    # Extract from alignment
    output_aln_name = args.out + '.aln'
    names_in_alignment = set()
    sequences_to_print = []
    alignment = AlignIO.read(args.aln,'fasta')
    for taxon in alignment:
        names_in_alignment.add(taxon.id)
        if taxon.id in subset:
            sequences_to_print.append(taxon)

    # Check subset sequences are found in alignment
    not_found_in_dataset = subset - names_in_alignment
    if len(not_found_in_dataset) > 0:
        sys.stderr.write('Sequences in subset missing from alignment: ' + \
                            str(not_found_in_dataset) + '\n')
        sys.exit(1)
    
    # Prune from the tree
    output_tree_name = args.out + '.tree'
    tree = dendropy.Tree.get(path = args.tree,
                             schema = 'newick',
                             preserve_underscores = True)
    tree.retain_taxa_with_labels(subset)
    tree.write_to_path(output_tree_name,
                        'newick')

    # Identify relevant recombination blocks
    recombination_starts = []
    recombination_ends = []
    recombination_taxa = []
    output_gff_name = args.out + '.gff'
    taxon_pattern = re.compile('taxa="([^"]*)"')
    with open(args.gff,'r') as in_gff, open(output_gff_name,'w') as out_gff:
        for line in in_gff.readlines():
            if line.startswith('##'):
                out_gff.write(line)
            else:
                info = line.rstrip().split('\t')
                taxon_set = set(taxon_pattern.search(info[8]).group(1).split())
                if not taxon_set.isdisjoint(subset):
                    out_gff.write(line)
                    if len(taxon_set) < len(subset):
                        recombination_starts.append(int(info[3]))
                        recombination_ends.append(int(info[4]))
                        recombination_taxa.append(frozenset(taxon_set))

    # Write out masked alignment for clade
    with open(output_aln_name,'w') as out_aln:
        for taxon in sequences_to_print:
            for index,taxa in enumerate(recombination_taxa):
                if taxon.id in taxa:
                    seq_string = str(taxon.seq)
                    replacement_string = args.missing_char*(recombination_ends[index] - recombination_starts[index] + 1)
                    seq_string = seq_string[:recombination_starts[index]-1] + replacement_string + seq_string[recombination_ends[index]:]
                    taxon.seq = Seq(seq_string)
            SeqIO.write(taxon, out_aln, args.out_fmt)



#! /usr/bin/env python3
# encoding: utf-8

"""
Tests the validation of the starting newick tree.
"""

import os
import unittest
from gubbins import common

modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestValidationOfStartingTree(unittest.TestCase):

    def test_are_all_names_unique(self):
        assert not common.is_starting_tree_valid(os.path.join(data_dir, 'duplicate_names_in_tree.tre'))

    def test_is_it_a_valid_newick_tree(self):
        assert not common.is_starting_tree_valid(os.path.join(data_dir, 'invalid_newick_tree.tre'))
        assert common.is_starting_tree_valid(os.path.join(data_dir, 'valid_newick_tree.tre'))

    def test_do_the_names_match_the_fasta_file(self):
        assert common.do_the_names_match_the_fasta_file(os.path.join(data_dir, 'valid_newick_tree.tre'),
                                                        os.path.join(data_dir, 'valid_newick_tree.aln'))


if __name__ == "__main__":
    unittest.main()



#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020 Imperial College London and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import sys
import os
import argparse
import subprocess
from multiprocessing import Pool
from functools import partial
from shutil import which

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Generate a ska2 alignment from a list '
                                                    'of assemblies',
                                     prog='generate_ska_alignment')

    # input options
    parser.add_argument('--reference',
                        help = 'Name of reference sequence to use for alignment',
                        required = True)
    parser.add_argument('--input',
                        help = 'List of sequence data; one row per isolate, with first column being the isolate name',
                        default = None,
                        required = False)
    parser.add_argument('--out',
                        help = 'Name of output alignment',
                        required = True)
    parser.add_argument('--k',
                        help = 'Split kmer size',
                        type = int,
                        default = 17)
    parser.add_argument('--threads',
                        help = 'Number of threads to use',
                        type = int,
                        default = 1)
    parser.add_argument('--no-cleanup',
                        help = 'Do not remove intermediate files',
                        action = 'store_true',
                        default = False)

    return parser.parse_args()

def map_fasta_sequence(seq, k = None, names = None):
    subprocess.check_output('ska fasta -o ' + names[seq]  + ' -k ' + str(k) + ' ' + seq,
                            shell = True)

def map_fastq_sequence(read_pair, k = None, names = None):
    subprocess.check_output('ska fastq -o ' + names[read_pair[0]]  + ' -k ' + str(k) + ' ' + \
                            read_pair[0] + ' ' + read_pair[1],
                            shell = True)

def ska_map_sequences(seq, k = None, ref = None):
    subprocess.check_output('ska map -o ' + seq + '.map -k ' + str(k) + ' -r ' + ref + \
                            ' ' + seq + '.skf',
                            shell = True)

# main code
if __name__ == "__main__":

    __spec__ = None

    # Get command line options
    args = get_options()

    # Check if ska is installed
    if which('ska') is None:
        sys.stderr.write('ska2 cannot be found on PATH; install with "conda install ska2"')
        sys.exit(1)

    # Check if k value is acceptable:
    if (args.k % 2) == 0 or args.k < 5 or args.k > 63:
        sys.stderr.write('k must be odd and between 5 and 63\n')
        sys.exit(1)

    # Build ska sketch
    subprocess.check_output('ska build -o ' + args.out + ' -k ' + str(args.k) + \
                            ' -f ' + args.input + ' --threads ' + str(args.threads),
                            shell = True)

    # Remove repetitive sequences
    if not os.path.exists(args.reference):
        sys.stderr.write('Reference file missing: ' + args.reference + '\n')
        sys.exit(1)

    # Run ska mapping
    if os.path.exists(args.out + '.skf'):
        tmp_aln = os.path.join(os.path.dirname(args.out), 'tmp.' + os.path.basename(args.out))
        subprocess.check_output('ska map -o ' + tmp_aln + ' --threads ' + str(args.threads) + ' --repeat-mask ' +  \
                                    args.reference + ' ' + args.out + '.skf',
                                shell = True)
    else:
        sys.stderr.write('ska building failed\n')
        sys.exit(1)
                            
    # Clean alignment to prep for Gubbins
    subprocess.check_output('gubbins_alignment_checker.py --aln ' + tmp_aln + ' --out-aln '  + args.out + \
                             ' --out ' + args.out + '.csv',
                            shell = True)
    
    sys.stderr.write("Completed generating alignment with ska2 (https://github.com/bacpop/ska.rust)\nPlease cite https://doi.org/10.1101/2024.03.25.586631\n")

    # Clean up
    if not args.no_cleanup:
        subprocess.check_output('rm ' + args.out + '.skf ' + tmp_aln,
                                shell = True)



#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2023  Wellcome Trust Sanger Institute and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import sys
import argparse
import re

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Mask recombinant regions detected by'
                                                    ' Gubbins from the input alignment',
                                     prog='mask_gubbins_aln')

    # input options
    parser.add_argument('--rec-gff',
                        help = 'GFF of recombinant regions detected by Gubbins',
                        required = True)
    parser.add_argument('--anno-gff',
                        help = 'GFF of annotation corresponding to the input alignment',
                        required = True)
    parser.add_argument('--out',
                        help = 'Output file name',
                        required = True)
    return parser.parse_args()

# main code
if __name__ == "__main__":

    # Get command line options
    args = get_options()
    
    # Read recombinant regions from GFF
    rec_start = []
    rec_end = []
    rec_affected = []
    taxon_pattern = re.compile('taxa="([^"]*)"')
    with open(args.rec_gff,'r') as gff_file:
        for line in gff_file.readlines():
            if not line.startswith('##'):
                # Calculate stats
                info = line.rstrip().split('\t')
                start = int(info[3])
                end = int(info[4])
                taxon_set = set(taxon_pattern.search(info[8]).group(1).split())
                # Record stats
                rec_start.append(start)
                rec_end.append(end)
                rec_affected.append(taxon_set)

    # Read annotation from GFF
    cds_start = []
    cds_end = []
    cds_name = []
    cds_index = []
    contig_starts = {}
    cumulative_length = 0
    with open(args.anno_gff,'r') as gff_file:
        for line in gff_file.readlines():
            if line.startswith('##sequence-region'):
                info = line.rstrip().split(' ')
                contig_starts[info[1]] = int(cumulative_length)
                cumulative_length = contig_starts[info[1]] + int(info[3])
            elif not line.startswith('##'):
                info = line.rstrip().split('\t')
                if len(info) >= 7:
                  if info[2] == 'CDS':
                      name = '-'
                      index = None
                      cds_data = info[8].replace('"','').split(';')
                      for datum in cds_data:
                          qualifier, value = datum.split('=')
                          if qualifier == 'locus_tag':
                              index = value
                          elif qualifier == 'gene':
                              name = value
                      if index is not None:
                          contig_start = contig_starts[info[0]]
                          cds_start.append(int(info[3]) + contig_start)
                          cds_end.append(int(info[4]) + contig_start)
                          cds_index.append(index)
                          cds_name.append(name)

    # Run checks on whether genes have been detected
    if len(cds_index) == 0:
        sys.stderr.write('No genes detected in annotation\n')
        sys.exit()
    elif not (len(cds_start) == len(cds_end) and len(cds_start) == len(cds_index) and len(cds_start) == len(cds_name)):
        sys.stderr.write('Error with extraction of information on annotation\n')
        sys.exit()
    elif not (len(rec_start) == len(rec_end) and len(rec_start) == len(rec_affected)):
        sys.stderr.write('Error with extraction of information on recombination\n')
        sys.exit()

    # Write out summary statistics
    with open(args.out,'w') as out_file:
        out_file.write('CDS\tGeneName\tStart\tEnd\tNumRec\tNumAffectedTaxa\tAffectedTaxa\n')
        for cnum,cindex in enumerate(cds_index):
            num_rec = 0
            affected_taxa = set()
            cstart = cds_start[cnum]
            cend = cds_end[cnum]
            for rnum,rstart in enumerate(rec_start):
                rend = rec_end[rnum]
                if (rstart < cend and rend > cend) or \
                   (rstart < cstart and rend > cstart) or \
                   (rstart > cstart and rend < cend):
                    num_rec = num_rec + 1
                    affected_taxa = affected_taxa.union(rec_affected[rnum])
            sorted_affected_taxa = list(affected_taxa)
            sorted_affected_taxa.sort()
            out_file.write(cds_index[cnum] + '\t' + cds_name[cnum] + '\t' + str(cds_start[cnum]) + '\t' + str(cds_end[cnum]) + '\t' + str(num_rec) + '\t' + str(len(affected_taxa)) + '\t' + ';'.join(sorted_affected_taxa) + '\n')
